@@ -97,8 +97,7 @@ GDALDataset *S104Dataset::Open(GDALOpenInfo *poOpenInfo)
     // Confirm the requested access is supported.
     if (poOpenInfo->eAccess == GA_Update)
     {
-        CPLError(CE_Failure, CPLE_NotSupported,
-                 "The S104 driver does not support update access.");
+        ReportUpdateNotSupportedByDriver("S104");
         return nullptr;
     }
 
