@@ -8,23 +8,7 @@
  ******************************************************************************
  * Copyright (c) 2002, Frank Warmerdam <warmerdam@pobox.com>
  *
- * Permission is hereby granted, free of charge, to any person obtaining a
- * copy of this software and associated documentation files (the "Software"),
- * to deal in the Software without restriction, including without limitation
- * the rights to use, copy, modify, merge, publish, distribute, sublicense,
- * and/or sell copies of the Software, and to permit persons to whom the
- * Software is furnished to do so, subject to the following conditions:
- *
- * The above copyright notice and this permission notice shall be included
- * in all copies or substantial portions of the Software.
- *
- * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
- * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
- * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
- * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
- * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
- * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
- * DEALINGS IN THE SOFTWARE.
+ * SPDX-License-Identifier: MIT
  ****************************************************************************/
 
 #include "ogr_oci.h"
@@ -334,9 +318,9 @@ int OGROCISession::EstablishSession(const char *pszUseridIn,
     OGROCIStatement oSetNLSTimeFormat(this);
     if (oSetNLSTimeFormat.Execute(
             "ALTER SESSION SET NLS_DATE_FORMAT='YYYY/MM/DD' \
-        NLS_TIME_FORMAT='HH24:MI:SS' NLS_TIME_TZ_FORMAT='HH24:MI:SS TZHTZM' \
-        NLS_TIMESTAMP_FORMAT='YYYY/MM/DD HH24:MI:SS' \
-        NLS_TIMESTAMP_TZ_FORMAT='YYYY/MM/DD HH24:MI:SS TZHTZM' \
+        NLS_TIME_FORMAT='HH24:MI:SS' NLS_TIME_TZ_FORMAT='HH24:MI:SS.FF TZHTZM' \
+        NLS_TIMESTAMP_FORMAT='YYYY/MM/DD HH24:MI:SS.FF' \
+        NLS_TIMESTAMP_TZ_FORMAT='YYYY/MM/DD HH24:MI:SS.FF TZHTZM' \
         NLS_NUMERIC_CHARACTERS = '. '") != CE_None)
         return OGRERR_FAILURE;
 
@@ -512,11 +496,16 @@ CPLErr OGROCISession::GetParamInfo(OCIParam *hParamDesc,
             poOGRDefn->SetType(OFTDate);
             break;
         case SQLT_TIMESTAMP:
+        case SQLT_TIME:
+            poOGRDefn->SetType(OFTDateTime);
+            break;
         case SQLT_TIMESTAMP_TZ:
         case SQLT_TIMESTAMP_LTZ:
-        case SQLT_TIME:
         case SQLT_TIME_TZ:
             poOGRDefn->SetType(OFTDateTime);
+            // Indicates that there's timezones. They might not actually be
+            // mixed !
+            poOGRDefn->SetTZFlag(OGR_TZFLAG_MIXED_TZ);
             break;
 
         case SQLT_RID:
